#ifndef _PV_H
#define _PV_H

#include <asm/byteorder.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include <ptxlist.h>

#define MAX_CONNECTIONS		10
#define MAX_BUFLEN		4096
#define MAX_DATA_LEN		(MAX_BUFLEN - sizeof (struct jvisu_sndvar))
#define MAX_STREAM_LEN		256
#define MAX_SYMBOL_LEN		256
#define JVISU_PORT		2200
#define JVISU_INTERVAL		20	/* time between variable checks in ms */

#define STAT_READ		1
#define STAT_WRITE		2

/* Network Commands are defined by the JVisu Native Protocol, so don't change them */
enum jvisu_cmd {
	cmd_alive		= 0,
	cmd_register_var	= 1,
	cmd_unregister_var	= 2,
	cmd_send_var		= 3,
	cmd_conf_data		= 4,
};

#define JVISU_STAT_INIT    0x01
#define JVISU_STAT_CHANGED 0x02

#define JVISU_PERM_RW 0x00
#define JVISU_PERM_RO 0x01
#define JVISU_PERM_WO 0x02

typedef enum {
	jv_double	= 1,
	jv_string	= 2,
	jv_bool		= 3,
	jv_byte		= 4,
	jv_short	= 5,
	jv_int		= 6,
	jv_float	= 7,
	jv_stream	= 8,
	jv_bytearray	= 9,
} jv_var_t;


/*
 * Byteorder stuff, stolen from linux/include/fs/jffs2.h
 */
typedef struct {
	uint64_t v64;
} __attribute__((packed))  jint64_t;

typedef struct {
	uint32_t v32;
} __attribute__((packed))  jint32_t;

typedef struct {
	uint16_t v16;
} __attribute__((packed)) jint16_t;

#define cpu_to_je16(x) ((jint16_t){__cpu_to_le16(x)})
#define cpu_to_je32(x) ((jint32_t){__cpu_to_le32(x)})
#define cpu_to_je64(x) ((jint64_t){__cpu_to_le64(x)})
#define je16_to_cpu(x) (__le16_to_cpu(x.v16))
#define je32_to_cpu(x) (__le32_to_cpu(x.v32))
#define je64_to_cpu(x) (__le64_to_cpu(x.v64))

/** jvisu command header */
struct jvisu_cmdhdr {
	u_char		cmd;
	u_char		type;
	jint16_t	len;
};

/** jvisu register variable command */
struct jvisu_regvar_cmd {
	struct jvisu_cmdhdr	hdr;
	jint32_t		idx;
	jint32_t		rate;
	u_char			symbol[0];	
};

/** jvisu stream variable */
struct jvisu_varstream {
	int	len;
	char	data[MAX_STREAM_LEN];
};

/** jvisu send variable command */
struct jvisu_sndvar {
	struct jvisu_cmdhdr	hdr;
	jint32_t		idx;
	union {
		u_char		b[0];
		jint16_t	s[0];
		jint32_t	i[0];
		jint64_t	d[0];
	} data;	
};

/** union to represent the internal variable of the application
 *  when writing 
 */
union jvisu_pv {
	u_char		b[0];
	short		s[0];
	int		i[0];
	float		f[0];
	double		d[0];
	uint16_t	u16[0];
	uint32_t	u32[0];
	uint64_t	u64[0];
};

/** jvisu generic telegram */
struct jvisu_sndgen {
	struct jvisu_cmdhdr	hdr;
	u_char			data[0];
};

/** socket structure */
struct jvisu_sock {
	int		sd;
	int		status;
	unsigned int	bufpoi;
	unsigned int	cmdlen;
	u_char		buf[MAX_BUFLEN];
};

/** jvisu registered variable */
struct jvisu_var {
//	struct list_head 	list;
	struct jvisu_sock	*sd;		/* socket descriptor */
	int			type;		/* changed by jvisu  */
	jv_var_t 		inttype;	/* internal type     */
	int			permissions;
	void 			(*write_done) (int arg);
	int			arg;
	int			idx;
	int			rate;
	int			elapsed;
	int			status;
	int			arraylen;
	void			*data;
	u_char			bitmask;
	u_char			dummy[3];
	union {
		u_char		b;
		uint16_t	s;
		uint32_t	i;
		float		f;
		double		d;
		u_char		*p;
	} cmp;
};

/** jvisu user supplied functions */
struct jvisu_ops {
	int 	(*get_symbol) (struct jvisu_var *var, char *symbol);
	int	(*save_configs) (int typ, char *data);
};

/* Macros for min/max.  */
#ifndef MIN
#define	MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef MAX
#define	MAX(a,b) (((a)>(b))?(a):(b))
#endif


/*
 *  Process Variable Engine
 */
#define XML_STR_BUF_SIZE 80
typedef struct {

	/* JVisu */
	int jv_port;			/* TCP port for JVisu; default=2202 */
	struct jvisu_sock jv_active_sock[MAX_CONNECTIONS + 1];
	int jv_connections; 
	fd_set jv_read_fdset;
	fd_set jv_write_fdset;
	ptxlist *jv_varlist; 
	int jv_registeredvars;		/* FIXME: replace */
	struct timeval jv_tact;
	
	/* XML PV Description */
	char *xml_filename;		/* XML descr. of process variables  */
	FILE *xml_file;
        int  xml_depth;
        char xml_str_buf[XML_STR_BUF_SIZE];
        char xml_str_buf_len;
        int  xml_in_tag;
        char *xml_pv_name_str;
        char *xml_pv_type_str;
        char *xml_pv_offset_str;
        char *xml_pv_desc_str;
        char *xml_pv_default_str;
        char *xml_pv_scale_str;
        char *xml_pv_unit_str;

	/* PV Backing Store */
	char *pv_filename;		/* process variable backing store   */
	int pv_file;
	int pv_sem;			/* SysV semaphore                   */
	int pv_sem_key;
	void *pv_map;			/* mmap area                        */
	
	/* PV Data */
	ptxlist *pv_list;		/* list of process variables        */

	/* JVisu */
	pthread_t jvsocket_thread;
	int jvisu_stop;			/* set 1 to stop JVisu server       */

} pv_engine_t;

/*
 *  Process Variable 
 */

enum pv_type {
	pv_type_undef = 0,
	pv_type_uint8,
	pv_type_uint16,
	pv_type_uint32,
};

enum pv_unit {
	pv_unit_none = 0,
};

enum pv_scale {
	pv_scale_none = 0,
	pv_scale_giga,
	pv_scale_mega,
	pv_scale_kilo,
	pv_scale_milli,
	pv_scale_micro,
	pv_scale_nano,
	pv_scale_pico,
	pv_scale_custom,
};

typedef struct {
	char 		*name;
	enum pv_type 	type;
	unsigned long 	offset;
	char		*desc;
	void 		*dflt;
	enum pv_scale	scale;
	enum pv_unit	unit;
	void		*data;
} pv_t; 


/*
 *  API functions
 */
extern pv_engine_t *pv_init(int argc, char *argv[]);
extern void         pv_finish (pv_engine_t *pv_engine);

extern uint8_t *pv_register_uint8(pv_engine_t *pv_engine, const char *varname);
extern uint16_t *pv_register_uint16(pv_engine_t *pv_engine, const char *varname);
extern uint32_t *pv_register_uint32(pv_engine_t *pv_engine, const char *varname);

extern int pv_stop_jvserver(pv_engine_t *pv_engine, int stop);
extern int pv_debuglevel(pv_engine_t *pv_engine, int level);
extern int pv_jvsocket(pv_engine_t *pv_engine, struct jvisu_ops *ops, int port);

static inline void pv_lock(pv_engine_t *pv_engine)
{
	static struct sembuf op = {0,-1,SEM_UNDO};
	if (semop(pv_engine->pv_sem, &op, 1) < 0) {
		perror("pv_lock");
		exit(EXIT_FAILURE);
	}
}

static inline void pv_unlock(pv_engine_t *pv_engine)
{
	static struct sembuf op = {0,1,SEM_UNDO};
	if (semop(pv_engine->pv_sem, &op, 1) < 0) {
		perror("pv_unlock");
		exit(EXIT_FAILURE);
	}
}

#endif
